#include "StdAfx.h"


#ifndef DISABLE_OPENNI


#ifdef  WIN64
#pragma comment(lib, "XnVNITE64_1_5_2.lib")
#pragma comment(lib, "OpenNI64.lib")
#else
#pragma comment(lib, "XnVNITE_1_5_2.lib")
#pragma comment(lib, "OpenNI.lib")
#endif





#include  "ExOpenNiWin.h"
#include  "WinTools.h"




/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// CExOpenNiWin NX
//

CExOpenNiWin::CExOpenNiWin()
{
	dev_backup		= NULL;
	niFileDev		= NULL;

	niJoints		= NULL;
	niNetwork		= NULL;
	sharedMem		= NULL;
	niExportData	= NULL;

	pLogFrame		= NULL;
	pSensorFrame	= NULL;
	pSkeletonFrame	= NULL;
	pLogDoc			= NULL;

	dataSaving		= FALSE;

	vect_fwrd.set(1.0, 0.0, 0,0, 1.0);

	appParam.init();
	tempFilePath = makeTempFilePath();
}




CExOpenNiWin::~CExOpenNiWin()
{
	if (dev_backup!=NULL) delete(dev_backup);

	dev_backup		= NULL;
	niJoints		= NULL;
	niNetwork		= NULL;
	sharedMem		= NULL;
	pLogFrame		= NULL;
	pSensorFrame	= NULL;
	pSkeletonFrame	= NULL;
	pLogDoc			= NULL;
}




///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//

CString  CExOpenNiWin::makeTempFilePath(void)
{		
	return MakeWorkingFolderPath(_T(OPENNI_RECORDE_TEMP_FILE), TRUE, _T("\\NSL\\Rinions\\"), TRUE);
}



void  CExOpenNiWin::setLogFramePtr(CExTextFrame* pfrm)
{
	pLogFrame = pfrm;
	
	if (pfrm!=NULL) pLogDoc = (CLogWndDoc*)(((CLogWndFrame*)pLogFrame)->pDoc);
	else            pLogDoc = NULL;
}



void  CExOpenNiWin::setParameter(CParameterSet param)
{ 
	appParam  = param;

	m_use_nite_smth = param.useNiteSmooth;
	m_confidence	= param.confidence;
	m_nite_smooth	= param.smoothNITE;
	m_profile		= param.detectParts;
}






///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//

BOOL  CExOpenNiWin::makeFileDevice(LPCTSTR fname)
{
	BOOL ret = FALSE;

	if (dev_backup!=NULL) {
		ret = remakeFileDevice(fname);
	}
	else {
		dev_backup = device;
		device = new COpenNiDevice();
		char* mbstr = ts2mbs(fname);
		ret = device->init(mbstr, TRUE);
		::free(mbstr);
		if (!ret) restoreDevice();
	}

	return ret;
}



BOOL  CExOpenNiWin::remakeFileDevice(LPCTSTR fname)
{
	device->delete_Depth();
	device->delete_Image();
	device->delete_Player();

	char* mbstr = ts2mbs(fname);
	BOOL ret = device->init(mbstr, TRUE);
	::free(mbstr);
	return ret;
}



BOOL  CExOpenNiWin::restoreDevice(void)
{
	if (dev_backup!=NULL) {
		delete(device);
		device = dev_backup;
		dev_backup = NULL;
	}

	return TRUE;
}






///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Virtual Function
//

void  CExOpenNiWin::saveJointsData(void)
{
	if (dataSaving && niFileDev!=NULL) {
		niFileDev->writeTempJText(posVect, rotQuat, m_is_mirroring);
	}
}



void  CExOpenNiWin::logingJointsData(void)
{
	if (!isNull(pLogDoc)) {
		pLogDoc->lock();
		for (int j=0; j<OPENNI_JOINT_NUM; j++) {
			XnVector3D  pos = jointPositionData(j);
			XnMatrix3X3 rot = jointRotationData(j);
			
			int n = j;
			if (m_is_mirroring) n = NiSDKMirrorJointNum(j, NiSDK_OpenNI);
			if (appParam.printPosMode || appParam.printRotMode || appParam.printQutMode) {
				if (pLogDoc!=NULL) pLogDoc->printFormat("LOCAL: %s (%2d)\n", NiSDK2JointName(n, NiSDK_OpenNI).c_str(), j);
			}
			
			if (appParam.printPosMode) {
				if (pLogDoc!=NULL) pLogDoc->printFormat("LOCAL: POS1: %f, %f, %f\n", pos.X, pos.Y, pos.Z);
				if (pLogDoc!=NULL) pLogDoc->printFormat("LOCAL: POS2: %f, %f, %f\n", posVect[n].x, posVect[n].y, posVect[n].z);
			}
			if (appParam.printRotMode) {
				if (pLogDoc!=NULL) pLogDoc->printFormat("LOCAL: ROT1: %f, %f, %f\n", rot.elements[0], rot.elements[1], rot.elements[2]);
				if (pLogDoc!=NULL) pLogDoc->printFormat("LOCAL: ROT2: %f, %f, %f\n", rot.elements[3], rot.elements[4], rot.elements[5]);
				if (pLogDoc!=NULL) pLogDoc->printFormat("LOCAL: ROT3: %f, %f, %f\n", rot.elements[6], rot.elements[7], rot.elements[8]);
			}
			if (appParam.printQutMode) {
				if (pLogDoc!=NULL) pLogDoc->printFormat("LOCAL: QUAT: %f, %f, %f, %f\n", rotQuat[n].x, rotQuat[n].y, rotQuat[n].z, rotQuat[n].s);
			}
		}
		if (!isNull(pLogDoc)) pLogDoc->unlock();
	}
}




void  CExOpenNiWin::convertJointsData(void)
{
	if (appParam.usePosData) {
		if (appParam.detectParts==XN_SKEL_PROFILE_ALL) {
			convertPos2JointsData();
		}
		else if (appParam.detectParts==XN_SKEL_PROFILE_UPPER) {
			convertUpperPos2JointsData();
		}
		else {
			return;
		}
	}
	else {
		convertRot2JointsData();
	}
	//
	if (pSensorFrame==NULL) return;

	if (niExportData!=NULL) {
		niExportData->exportJointsData(posVect, rotQuat, niNetwork, NiSDK_OpenNI, OPENNI_JOINT_NUM);
	}
}





void  CExOpenNiWin::convertPos2JointsData(void)
{
	//
	for (int j=0; j<OPENNI_JOINT_NUM; j++) {
		if (posVect[j].c<m_confidence) {
			Vector<double>* ptr = (Vector<double>*)posRing[j].get(-1);
			if (ptr!=NULL && ptr->c>0.0) posVect[j] = *ptr;
			else posVect[j].init(-1.0);
		}
	}

	//
	if (appParam.useMvavSmooth) niJoints->PosMovingAverage(NiSDK_OpenNI);
	if (appParam.useJointConst) niJoints->PosVibNoiseCanceler();


	//
	Vector<double> torso_down  = posVect[NI_SDK_PELVIS]  - posVect[NI_SDK_TORSO];
	Vector<double> shldr_left  = posVect[NI_SDK_L_SHLDR] - posVect[NI_SDK_R_SHLDR];
	Vector<double> torso_up	   = - torso_down;
	Vector<double> shldr_right = - shldr_left;

	// PELVIS YZʂ̉]
	if (posVect[NI_SDK_R_HIP].c>=m_confidence && posVect[NI_SDK_L_HIP].c>=m_confidence) { 
		Vector<double> vect = NewellMethod4(posVect[NI_SDK_R_HIP], posVect[NI_SDK_PELVIS], posVect[NI_SDK_L_HIP], posVect[NI_SDK_TORSO]);
		
		rotQuat[NI_SDK_PELVIS] = V2VQuaternion(vect_fwrd, vect);
		if (rotQuat[NI_SDK_PELVIS].c==-1.0 && rotQuat[NI_SDK_PELVIS].s==0.0 && rotQuat[NI_SDK_PELVIS].n==0.0) {
			rotQuat[NI_SDK_PELVIS].setRotation(PI, torso_up);
		}
	}
	else {
		rotQuat[NI_SDK_PELVIS].init();
	}

	// X ]
	Quaternion quat_lhip = ~rotQuat[NI_SDK_PELVIS]*posVect[NI_SDK_L_HIP]*rotQuat[NI_SDK_PELVIS];
	Quaternion quat_rhip = ~rotQuat[NI_SDK_PELVIS]*posVect[NI_SDK_R_HIP]*rotQuat[NI_SDK_PELVIS];

	Vector<double> hip_left  = quat_lhip.getVector() - quat_rhip.getVector(); 
	Vector<double> hip_right = - hip_left;
	double thx = atan2(hip_left.z, hip_left.y);
	Quaternion xrot;
	xrot.setRotation(thx, 1.0, 0.0, 0.0, 1.0);

	// Y]␳
	Quaternion yrot;
	yrot.setRotation((double)appParam.YaxisCorrect, 0.0, 1.0, 0.0, 1.0);

	//
	rotQuat[NI_SDK_PELVIS] = rotQuat[NI_SDK_PELVIS]*xrot*yrot;

	//
	rotQuat[NI_SDK_TORSO]  = VPPQuaternion(torso_up, posVect[NI_SDK_TORSO], posVect[NI_SDK_NECK]);
	rotQuat[NI_SDK_NECK]   = PPPQuaternion(posVect[NI_SDK_TORSO],  posVect[NI_SDK_NECK],  posVect[NI_SDK_HEAD]);

	// To Relative Coordinate
	rotQuat[NI_SDK_L_SHLDR] = VPPQuaternion(shldr_left,  posVect[NI_SDK_L_SHLDR], posVect[NI_SDK_L_ELBOW]);
	rotQuat[NI_SDK_R_SHLDR] = VPPQuaternion(shldr_right, posVect[NI_SDK_R_SHLDR], posVect[NI_SDK_R_ELBOW]);
	rotQuat[NI_SDK_L_HIP]	= VPPQuaternion(torso_down,	 posVect[NI_SDK_L_HIP],   posVect[NI_SDK_L_KNEE]);
	rotQuat[NI_SDK_R_HIP]	= VPPQuaternion(torso_down,  posVect[NI_SDK_R_HIP],   posVect[NI_SDK_R_KNEE]);
	//
	rotQuat[NI_SDK_L_ELBOW] = PPPQuaternion(posVect[NI_SDK_L_SHLDR], posVect[NI_SDK_L_ELBOW], posVect[NI_SDK_L_HAND]);
	rotQuat[NI_SDK_R_ELBOW] = PPPQuaternion(posVect[NI_SDK_R_SHLDR], posVect[NI_SDK_R_ELBOW], posVect[NI_SDK_R_HAND]);
	rotQuat[NI_SDK_L_KNEE]	= PPPQuaternion(posVect[NI_SDK_L_HIP],	 posVect[NI_SDK_L_KNEE],  posVect[NI_SDK_L_FOOT]);
	rotQuat[NI_SDK_R_KNEE]	= PPPQuaternion(posVect[NI_SDK_R_HIP],	 posVect[NI_SDK_R_KNEE],  posVect[NI_SDK_R_FOOT]);

	// To Avatar Coordinate
	rotQuat[NI_SDK_L_ELBOW] = ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_L_ELBOW]*rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_R_ELBOW] = ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_R_ELBOW]*rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_L_SHLDR] = ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_L_SHLDR]*rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_R_SHLDR] = ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_R_SHLDR]*rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_L_KNEE]	= ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_L_KNEE] *rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_R_KNEE]	= ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_R_KNEE] *rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_L_HIP]	= ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_L_HIP]	 *rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_R_HIP]	= ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_R_HIP]	 *rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_NECK]	= ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_NECK]	 *rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_TORSO]   = ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_TORSO]	 *rotQuat[NI_SDK_PELVIS];


	//
	for (int j=0; j<OPENNI_JOINT_NUM; j++) {
		if (rotQuat[j].c<m_confidence) {
			Quaternion* ptr = (Quaternion*)rotRing[j].get(-1);
			if (ptr!=NULL && ptr->c>0.0) rotQuat[j] = *ptr;
			else rotQuat[j].init();
		}
	}

	//
	if (appParam.useJointConst) niJoints->CheckJointRotation();
}





void  CExOpenNiWin::convertUpperPos2JointsData(void)
{
	//
	for (int j=0; j<OPENNI_JOINT_NUM; j++) {
		if (posVect[j].c<m_confidence) {
			Vector<double>* ptr = (Vector<double>*)posRing[j].get(-1);
			if (ptr!=NULL && ptr->c>0.0) posVect[j] = *ptr;
			else posVect[j].init(-1.0);
		}
	}

	//
	if (appParam.useMvavSmooth) niJoints->PosMovingAverage(NiSDK_OpenNI);
	if (appParam.useJointConst) niJoints->PosVibNoiseCanceler();


	//
	Vector<double> torso_down  = posVect[NI_SDK_PELVIS]  - posVect[NI_SDK_TORSO];
	Vector<double> shldr_left  = posVect[NI_SDK_L_SHLDR] - posVect[NI_SDK_R_SHLDR];
	Vector<double> torso_up	   = - torso_down;
	Vector<double> shldr_right = - shldr_left;


	// PELVIS YZʂ̉]
	if (posVect[NI_SDK_R_SHLDR].c>=m_confidence && posVect[NI_SDK_L_SHLDR].c>=m_confidence) { 
		Vector<double> vect = NewellMethod4(posVect[NI_SDK_R_SHLDR], posVect[NI_SDK_TORSO], posVect[NI_SDK_L_SHLDR], posVect[NI_SDK_NECK]);
		
		rotQuat[NI_SDK_PELVIS] = V2VQuaternion(vect_fwrd, vect);
		if (rotQuat[NI_SDK_PELVIS].c==-1.0 && rotQuat[NI_SDK_PELVIS].s==0.0 && rotQuat[NI_SDK_PELVIS].n==0.0) {
			rotQuat[NI_SDK_PELVIS].setRotation(PI, torso_up);
		}
	}
	else {
		rotQuat[NI_SDK_PELVIS].init();
	}

	// X ]
	Quaternion quat_lshd = ~rotQuat[NI_SDK_PELVIS]*posVect[NI_SDK_L_SHLDR]*rotQuat[NI_SDK_PELVIS];
	Quaternion quat_rshd = ~rotQuat[NI_SDK_PELVIS]*posVect[NI_SDK_R_SHLDR]*rotQuat[NI_SDK_PELVIS];

	Vector<double> shd_left  = quat_lshd.getVector() - quat_rshd.getVector(); 
	Vector<double> shd_right = - shd_left;
	double thx = atan2(shd_left.z, shd_left.y);
	Quaternion xrot;
	xrot.setRotation(thx, 1.0, 0.0, 0.0, 1.0);

	// Y]␳
	Quaternion yrot;
	yrot.setRotation((double)appParam.YaxisCorrect, 0.0, 1.0, 0.0, 1.0);

	//
	rotQuat[NI_SDK_PELVIS] = rotQuat[NI_SDK_PELVIS]*xrot*yrot;

	//
	rotQuat[NI_SDK_TORSO]  = VPPQuaternion(torso_up, posVect[NI_SDK_TORSO], posVect[NI_SDK_NECK]);
	rotQuat[NI_SDK_NECK]   = PPPQuaternion(posVect[NI_SDK_TORSO],  posVect[NI_SDK_NECK],  posVect[NI_SDK_HEAD]);

	// To Relative Coordinate
	rotQuat[NI_SDK_L_SHLDR] = VPPQuaternion(shldr_left,  posVect[NI_SDK_L_SHLDR], posVect[NI_SDK_L_ELBOW]);
	rotQuat[NI_SDK_R_SHLDR] = VPPQuaternion(shldr_right, posVect[NI_SDK_R_SHLDR], posVect[NI_SDK_R_ELBOW]);
	//
	rotQuat[NI_SDK_L_ELBOW] = PPPQuaternion(posVect[NI_SDK_L_SHLDR], posVect[NI_SDK_L_ELBOW], posVect[NI_SDK_L_HAND]);
	rotQuat[NI_SDK_R_ELBOW] = PPPQuaternion(posVect[NI_SDK_R_SHLDR], posVect[NI_SDK_R_ELBOW], posVect[NI_SDK_R_HAND]);

	//
	Quaternion pelvisXY;
	Vector<double> euler = Quaternion2EulerXYZ(rotQuat[NI_SDK_PELVIS]);
	euler.z = 0.0;
	pelvisXY.setEulerXYZ(euler);

	// To Avatar Coordinate
	rotQuat[NI_SDK_L_ELBOW] = ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_L_ELBOW]*rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_R_ELBOW] = ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_R_ELBOW]*rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_L_SHLDR] = ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_L_SHLDR]*rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_R_SHLDR] = ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_R_SHLDR]*rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_NECK]	= ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_NECK]	 *rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_TORSO]   = ~rotQuat[NI_SDK_PELVIS]*rotQuat[NI_SDK_TORSO]	 *rotQuat[NI_SDK_PELVIS];

	//
	rotQuat[NI_SDK_L_KNEE].init();
	rotQuat[NI_SDK_R_KNEE].init();
	rotQuat[NI_SDK_R_HIP]   = ~pelvisXY;
	rotQuat[NI_SDK_L_HIP]   = ~pelvisXY;


	//
	for (int j=0; j<OPENNI_JOINT_NUM; j++) {
		if (rotQuat[j].c<m_confidence) {
			Quaternion* ptr = (Quaternion*)rotRing[j].get(-1);
			if (ptr!=NULL && ptr->c>0.0) rotQuat[j] = *ptr;
			else rotQuat[j].init();
		}
	}

	//
	if (appParam.useJointConst) niJoints->CheckJointRotation();
}





void  CExOpenNiWin::convertRot2JointsData(void)
{
	//
	if (posVect[NI_SDK_R_HIP].c<m_confidence) {
		Vector<double>* ptr = (Vector<double>*)posRing[NI_SDK_R_HIP].get(-1);
		if (ptr!=NULL && ptr->c>0.0) posVect[NI_SDK_R_HIP] = *ptr;
		else posVect[NI_SDK_R_HIP].init(-1.0);
	}
	if (posVect[NI_SDK_L_HIP].c<m_confidence) {
		Vector<double>* ptr = (Vector<double>*)posRing[NI_SDK_L_HIP].get(-1);
		if (ptr!=NULL && ptr->c>0.0) posVect[NI_SDK_L_HIP] = *ptr;
		else posVect[NI_SDK_L_HIP].init(-1.0);
	}

	//
	Vector<double> torso_down  = posVect[NI_SDK_PELVIS]  - posVect[NI_SDK_TORSO];
	Vector<double> shldr_left  = posVect[NI_SDK_L_SHLDR] - posVect[NI_SDK_R_SHLDR];
	Vector<double> torso_up	   = - torso_down;
	Vector<double> shldr_right = - shldr_left;

	// PELVIS YZʂ̉]
	if (posVect[NI_SDK_R_HIP].c>=m_confidence && posVect[NI_SDK_L_HIP].c>=m_confidence) { 
		Vector<double> vect = NewellMethod4(posVect[NI_SDK_R_HIP], posVect[NI_SDK_PELVIS], posVect[NI_SDK_L_HIP], posVect[NI_SDK_TORSO]);
		
		rotQuat[NI_SDK_PELVIS]  = V2VQuaternion(vect_fwrd, vect);
		if (rotQuat[NI_SDK_PELVIS].c==-1.0 && rotQuat[NI_SDK_PELVIS].s==0.0 && rotQuat[NI_SDK_PELVIS].n==0.0) {
			rotQuat[NI_SDK_PELVIS].setRotation(PI, torso_up);
		}
	}
	else {
		rotQuat[NI_SDK_PELVIS].init();
	}

	// X ]iWϊj
	Quaternion quat_lhip = ~rotQuat[NI_SDK_PELVIS]*posVect[NI_SDK_L_HIP]*rotQuat[NI_SDK_PELVIS];
	Quaternion quat_rhip = ~rotQuat[NI_SDK_PELVIS]*posVect[NI_SDK_R_HIP]*rotQuat[NI_SDK_PELVIS];

	Vector<double> hip_left = quat_lhip.getVector() - quat_rhip.getVector(); 
	Vector<double> hip_right = - hip_left;
	double thx = atan2(hip_left.z, hip_left.y);
	Quaternion xrot;
	xrot.setRotation(thx, 1.0, 0.0, 0.0, 1.0);

	//
	rotQuat[NI_SDK_PELVIS]  = rotQuat[NI_SDK_PELVIS]*xrot;
	
	//
	rotQuat[NI_SDK_R_KNEE]  = rotQuat[NI_SDK_R_KNEE] *~rotQuat[NI_SDK_R_HIP];
	rotQuat[NI_SDK_L_KNEE]  = rotQuat[NI_SDK_L_KNEE] *~rotQuat[NI_SDK_L_HIP];
	rotQuat[NI_SDK_R_ELBOW] = rotQuat[NI_SDK_R_ELBOW]*~rotQuat[NI_SDK_R_SHLDR];
	rotQuat[NI_SDK_L_ELBOW] = rotQuat[NI_SDK_L_ELBOW]*~rotQuat[NI_SDK_L_SHLDR];

	rotQuat[NI_SDK_R_SHLDR] = rotQuat[NI_SDK_R_SHLDR]*~rotQuat[NI_SDK_NECK];
	rotQuat[NI_SDK_L_SHLDR] = rotQuat[NI_SDK_L_SHLDR]*~rotQuat[NI_SDK_NECK];
	rotQuat[NI_SDK_R_HIP]   = rotQuat[NI_SDK_R_HIP]  *~rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_L_HIP]   = rotQuat[NI_SDK_L_HIP]  *~rotQuat[NI_SDK_PELVIS];
	rotQuat[NI_SDK_NECK]    = rotQuat[NI_SDK_NECK]   *~rotQuat[NI_SDK_TORSO];
	rotQuat[NI_SDK_TORSO]   = rotQuat[NI_SDK_TORSO]  *~rotQuat[NI_SDK_PELVIS];
	

	// Confidence
	for (int j=0; j<OPENNI_JOINT_NUM; j++) {
		if (rotQuat[j].c<m_confidence) {
			Quaternion* ptr = (Quaternion*)rotRing[j].get(-1);
			if (ptr!=NULL && ptr->c>0.0) rotQuat[j] = *ptr;
			else rotQuat[j].init();
		}
	}

	//	
	if (appParam.useMvavSmooth) niJoints->RotMovingAverage(NiSDK_OpenNI);
	if (appParam.useJointConst) {
		niJoints->RotVibNoiseCanceler();
		niJoints->CheckJointRotation();
	}
}



void  CExOpenNiWin::lostTrackingUser(int uid)
{ 
	DEBUG_ERR("LOST   TRACKING USER (%d)", uid);

	clearJointsData();
	//
	sharedMem->isTracking = FALSE;
	sharedMem->clearLocalAnimationData();

	if (niNetwork->sendSocket>0) {
		if (appParam.netOutMode==NETandLOCAL) sharedMem->clearLocalAnimationIndex();
		//
		if (appParam.netFastMode) {
			sendAnimationData<float> (posVect, rotQuat, niNetwork, NiSDK_OpenNI, OPENNI_JOINT_NUM);
		}
		else {
			sendAnimationData<double>(posVect, rotQuat, niNetwork, NiSDK_OpenNI, OPENNI_JOINT_NUM);
		}
	}
}




void  CExOpenNiWin::detectTrackingUser(int uid)
{ 
	DEBUG_ERR("DETECT TRACKING USER (%d)", uid);

	sharedMem->isTracking = TRUE;
}



//////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// for TEST
//

bool  CExOpenNiWin::setStartBoneLength()
{
	bool ret = false;

	memset(stbnLen, 0, sizeof(double)*OPENNI_JOINT_NUM);

	if (stbnLen[ 1]==0.0 && posVect[ 1].n!=0.0 && posVect[ 2].n!=0.0) {
		stbnLen[ 1] = VectorDist(posVect[ 1], posVect[ 2]);
	}
	if (stbnLen[ 2]==0.0 && posVect[ 2].n!=0.0 && posVect[ 3].n!=0.0) {
		stbnLen[ 2] = VectorDist(posVect[ 2], posVect[ 3]);
	}
	if (stbnLen[ 3]==0.0 && posVect[ 3].n!=0.0 && posVect[ 0].n!=0.0) {
		stbnLen[ 3] = VectorDist(posVect[ 3], posVect[ 0]);
	}

	if (stbnLen[ 6]==0.0 && posVect[ 6].n!=0.0 && posVect[ 2].n!=0.0) {
		stbnLen[ 6] = VectorDist(posVect[ 6], posVect[ 2]);
	}
	if (stbnLen[ 7]==0.0 && posVect[ 7].n!=0.0 && posVect[ 6].n!=0.0) {
		stbnLen[ 7] = VectorDist(posVect[ 7], posVect[ 6]);
	}
	if (stbnLen[ 9]==0.0 && posVect[ 9].n!=0.0 && posVect[ 7].n!=0.0) {
		stbnLen[ 9] = VectorDist(posVect[ 9], posVect[ 7]);
	}

	if (stbnLen[12]==0.0 && posVect[12].n!=0.0 && posVect[ 2].n!=0.0) {
		stbnLen[12] = VectorDist(posVect[12], posVect[ 2]);
	}
	if (stbnLen[13]==0.0 && posVect[13].n!=0.0 && posVect[12].n!=0.0) {
		stbnLen[13] = VectorDist(posVect[13], posVect[12]);
	}
	if (stbnLen[15]==0.0 && posVect[15].n!=0.0 && posVect[13].n!=0.0) {
		stbnLen[15] = VectorDist(posVect[15], posVect[13]);
	}

	if (stbnLen[17]==0.0 && posVect[17].n!=0.0 && posVect[ 0].n!=0.0) {
		stbnLen[17] = VectorDist(posVect[17], posVect[ 0]);
	}
	if (stbnLen[18]==0.0 && posVect[18].n!=0.0 && posVect[17].n!=0.0) {
		stbnLen[18] = VectorDist(posVect[18], posVect[17]);
	}
	if (stbnLen[20]==0.0 && posVect[20].n!=0.0 && posVect[18].n!=0.0) {
		stbnLen[20] = VectorDist(posVect[20], posVect[18]);
	}

	if (stbnLen[21]==0.0 && posVect[21].n!=0.0 && posVect[ 0].n!=0.0) {
		stbnLen[21] = VectorDist(posVect[21], posVect[ 0]);
	}
	if (stbnLen[22]==0.0 && posVect[22].n!=0.0 && posVect[21].n!=0.0) {
		stbnLen[22] = VectorDist(posVect[22], posVect[21]);
	}
	if (stbnLen[24]==0.0 && posVect[24].n!=0.0 && posVect[22].n!=0.0) {
		stbnLen[24] = VectorDist(posVect[24], posVect[22]);
	}

	return ret;
}








////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Xbh
//

UINT  openniEventLoop(LPVOID pParam)
{
	if (pParam==NULL) return 1;

	CExOpenNiWin* openni = (CExOpenNiWin*)pParam;
	if (openni->device->context==NULL) return 1; 
	if (openni->pSensorFrame==NULL)    return 1; 
	
	CExView* pview = openni->pSensorFrame->pView;
	openni->pViewData = &pview->viewData;
	
	//
	try {
		Loop {
			//
			if (openni->getDevState()==NI_STATE_DETECT_STOPPING || openni->getDevState()==NI_STATE_SAVE_WORKING) continue;
			openni->device->context->WaitAndUpdateAll();		
			
			if (openni->device->image!=NULL && openni->device->imageMD!=NULL && openni->m_use_image) {
				//openni->device->context->WaitOneUpdateAll(*(openni->device->image));
				openni->device->image->GetMetaData(*(openni->device->imageMD));
			}

			if (openni->getDevState()==NI_STATE_DETECT_STOPPING || openni->getDevState()==NI_STATE_SAVE_WORKING) continue;
			//openni->device->context->WaitOneUpdateAll(*(openni->device->depth));
			openni->hasDepthData = FALSE;
			if (openni->getDevState()==NI_STATE_DETECT_EXEC && openni->device->user!=NULL) {
				//openni->device->context->WaitOneUpdateAll(*(openni->device->user));
				openni->device->user->GetUserPixels(0, *openni->device->sceneMD);
				openni->hasDepthData = TRUE;
			}

			if (openni->getDevState()==NI_STATE_DETECT_STOPPING || openni->getDevState()==NI_STATE_SAVE_WORKING) continue;
			if (isNull(openni->pSensorFrame)) break;
			openni->makeDisplayImage();			// need Depth Data when detected users are painted

			if (openni->getDevState()==NI_STATE_DETECT_EXEC) {
				openni->trackingJoints(openni->appParam.useRotData);	// need Depth Data
			}

			//
			if (isNull(openni->pSensorFrame)) break;
			if (isNull(openni->pSensorFrame->pView)) break;
			if (!pview->SetNewSurface())    break;

			//
			if (isNull(openni->pSensorFrame)) break;
			pview->ExecRender();
		}
	}

	catch (std::exception& ex) {
		openni->m_err_mesg = _T("openniEventLoop() Exception: ");
		openni->m_err_mesg+= mbs2ts((char*)ex.what());
		return 2;
	}

	return 0;
}


#endif